﻿using System;
using System.Linq;
using System.Linq.Expressions;

static class Program
{
    static void Main()
    {
        // クエリ対象となるデータソース
        string[] dataSource = { "apple", "orange", "peach", "melon", "grape" };

        // 検索キーワードを3つ取得
        Console.WriteLine("指定した文字を含む文字列を検索します。");
        ConsoleKeyInfo info = Console.ReadKey();
        string character = info.KeyChar.ToString();
        Console.WriteLine("かlかoを含むものを検索。");
        string[] keywords = { character, "l", "o" };

        // メソッド・ベースのLINQ文（静的なラムダ式を利用）
        //var query = dataSource.AsQueryable().
        //    Where(item =>
        //        item.ToLower().Contains(keywords[0].ToLower()) ||
        //        item.ToLower().Contains(keywords[1].ToLower()) ||
        //        item.ToLower().Contains(keywords[2].ToLower())).
        //    OrderByDescending(item => item);

        // メソッド・ベースのLINQ文（動的な式ツリー型を利用）
        var query = dataSource.AsQueryable().
          Where(GetExpressionTreeWhere(keywords)).
          OrderByDescending(item => item);

        // 検索された用語を出力
        foreach (string term in query)
        {
            Console.WriteLine(term);
        }
        // 出力例（「m」を入力した場合）：
        // 指定した文字を含む文字列を検索します。
        // mかlかoを含むものを検索。
        // orange
        // melon
        // apple

        // 出力を確認するために実行を止める
        Console.ReadKey();
    }

    private static Expression<Func<string, bool>> GetExpressionTreeWhere(string[] keywords)
    {
        // 引数が適切でない場合には、例外を発行する
        int length = keywords.Length;
        if (length == 0)
        {
            throw new ArgumentException("キーワード指定なし。");
        }

        // ラムダ式における左辺のパラメータの名前
        // （「item => ……」の部分に対応）
        const string paramName = "item";

        // ラムダ式の左辺を構成するパラメータ項目の1つを作成
        ParameterExpression parameter =
          LambdaUtil.GetStringParameterExpression(paramName);

        // ラムダ式の左辺であるパラメータ（全項目）を配列にまとめる
        // （※ただし今回はパラメータ項目は1つしかない）
        ParameterExpression[] parameters =
          LambdaUtil.GetParameterExpressions(parameter);

        // ラムダ式の右辺である式を組み立てる
        Expression body = null;
        for (int i = 0; i < length; i++)
        {
            // 「item.ToLower().Contains(keywords[0].ToLower())」
            // という式を作成して、2つ目以降のキーワードは
            // 条件OR演算子（||）で連結する
            body = LambdaUtil.GetContainsExpression(
              parameter, keywords[i], (i == 0) ? null : body);
        }

        // 最後に、動的に作成したラムダ式から式ツリーを取得する
        return LambdaUtil.GetLambdaExpressionWhere(parameters, body);
    }

}
